//
//------------------------------------------------------------------------------
// Name    : HPWM16.bas
// Author  : Jerry Messina
// Date    : 05 DEC 2023
// Version : 1.00
//
// PWM module support for 16-bit PWM peripheral (Q family)
//
// ver 1.00 05 DEC 2023
//  - initial version, based loosely on HPWM.bas
//------------------------------------------------------------------------------
//
module HPWM16

// 
// The 16-bit PWM modules offer multiple outputs, programmable polarity, and a compare feature.
// Each module uses an independent timer source and contains a single output slice comprised of 
// up to two PWM outputs. Both outputs have independent duty cycles, but share the same operating mode. 
// Outputs must be assigned using PPS as there are no fixed output pins.
// 
// There are a number of PWM clock sources available, selected via the PWMxCLK register.
// The PWM source clock frequency can be reduced with the PWMxCPRE clock prescaler register, 
// with prescaler selections from 1:1 to 1:256.
//
// The PWM period is controlled via the PWMxPR 16-bit register pair PWMxPRH:PWMxPRL
//  PWM Period = (PWMxPR + 1)/fPWMxCLK
//      where fPWMxCLK = fPWM_clock_source/prescaler
//  PWM Freq = 1/PWM Period = fPWMxCLK/(PWMxPR + 1)
//  so, PWM Freq = fPWM_clock_source/((PWMxPR + 1)*prescaler) 
//
// There are six operating modes: Left Aligned, Right Aligned, Center Aligned, Variable Aligned, 
// Pulsed Compare, and Toggled Compare. The left and right aligned modes can also be operated in 
// Push-Pull mode, where the outputs are active every other PWM period. 
// Duty cycle settings are typically controlled via the PWMxS1P1 and PWMxS1P2 registers, but usage 
// varies depending on the operating mode. See the datasheet, PWM Tips and Tricks (DS400002158A)
// and AN3381 for a discussion and examples.
//
//  Nomenclature:
//  PWMxSaPn
//      x = PWM module number (current devices have from 2-4 PWM modules)
//      a = slice number (only 1 slice in all current devices)
//      n = output number, 1-2 (in most modes each slice can have two outputs)
//

#if (_pwm = 0)      // '#const _pwm' setting from device file
  #error _device + " does not support the PWM peripheral module"
#endif

// number of PWM peripheral modules supported by this file (default='#const _pwm' setting)
// this can be used to reduce code size if all pwm modules are not required
#option HPWM16_MAX_PWM = _pwm       // default to max number from the device file
#if (HPWM16_MAX_PWM > _pwm)
  #error HPWM16_MAX_PWM, "HPWM16_MAX_PWM cannot exceed device file '#const _pwm' setting" 
#endif
const MAX_PWM = HPWM16_MAX_PWM

// set pwm clock source to FOSC during initialization
#option HPWM16_SET_DEFAULT_CLOCKSRC = true
#if not(HPWM16_SET_DEFAULT_CLOCKSRC in (true, false))
  #error HPWM16_SET_DEFAULT_CLOCKSRC, "invalid HPWM16_SET_DEFAULT_CLOCKSRC setting"
#endif

// default setting for autoload variable (SetPWMx, SetFreq)
// if this is false then you'll have to call Load() 
#option HPWM16_DEFAULT_AUTOLOAD = true
#if not(HPWM16_DEFAULT_AUTOLOAD in (true, false))
  #error HPWM16_DEFAULT_AUTOLOAD, "invalid HPWM16_DEFAULT_AUTOLOAD setting"
#endif

// default duty cycle for SetFreq
#option HPWM16_DEFAULT_DUTYCYCLE = 50
#if (HPWM16_DEFAULT_DUTYCYCLE > 100)
  #error HPWM16_DEFAULT_DUTYCYCLE, "invalid HPWM16_DEFAULT_DUTYCYCLE setting"
#endif
const DEFAULT_DUTYCYCLE = HPWM16_DEFAULT_DUTYCYCLE

// find closest freq
// if option HPWM16_CLOSEST_FREQ = true then SetFreq will try to find
//  the closest match to the requested pFreq
// otherwise it will use the freq >= requested value (which may be MUCH larger)
// adds ~250 bytes of code
#option HPWM16_CLOSEST_FREQ = false
#if not(HPWM16_CLOSEST_FREQ in (true, false))
  #error HPWM16_CLOSEST_FREQ, "invalid HPWM16_CLOSEST_FREQ setting"
#endif

//------------------------------------------------------------------------------
// PWMx peripheral registers
//  PWMxERS     PWMx External Reset Source (not used)
//  PWMxCLK     PWMx Clock Source
//  PWMxLDS     PWMx Auto-load Trigger Source Select Register (not used)
//  PWMxPR      PWMx Period Register
//  PWMxCPRE    PWMx Clock Prescaler Register
//  PWMxPIPOS   PWMx Period Interrupt Postscaler Register (not used)
//  PWMxGIR     PWMx Interrupt Register (not used)
//  PWMxGIE     PWMx Interrupt Enable Register (not used)
//  PWMxCON     PWMx Control Register
//  PWMxS1CFG   PWM Slice 1 Configuration Register
//  PWMxS1P1    PWM Slice 1 Parameter 1 Register (output 1)
//  PWMxS1P2    PWM Slice 1 Parameter 2 Register (output 2)
//  PWMLOAD     Mirror copies of all PWMxLD bits
//  PWMEN       Mirror copies of all PWMxEN bits

// PWMxCLK PWMx Clock Source Select Register (see datasheet for others)
public const
    PWMxCLK_FOSC            = %00010,
    PWMxCLK_HFINTOSC        = %00011,
    PWMxCLK_LFINTOSC        = %00100,
    PWMxCLK_MFINTOSC_500K   = %00101,
    PWMxCLK_MFINTOSC_31K    = %00110

// PWMxCON PWMx Control Register bits
public const
    EN = 7,         // PWM Module Enable
    LD = 2,         // Reload Registers
    ERSPOL = 1,     // External Reset Polarity Select
    ERSNOW = 0      // External Reset Mode Select

// PWMxS1CFG PWMx Configuration Register bits
public const
    POL2 = 7,       // pwm output 2 polarity, 0=high true, 1=low true
    POL1 = 6,       // pwm output 1 polarity, 0=high true, 1=low true
    PPEN = 3,       // Push-Pull Mode Enable (alternates out1->out2)
    MODE2 = 2,      // PWM Module Operating Mode Select MODE[2:0]
    MODE1 = 1,
    MODE0 = 0

// PWMxS1CFG MODE[2:0] operating mode
public const
    PWMxCFG_LEFT_ALIGN = %000,          // two outputs, active pulse at beginning
    PWMxCFG_RIGHT_ALIGN = %001,         // two outputs, active pulse at end
    PWMxCFG_CENTER_ALIGN = %010,        // two outputs, active center, period is 2x PWMxPR register
    PWMxCFG_VAR_ALIGN = %011,           // param1 defines start, param2 defines end, output1=output2
    PWMxCFG_COMPARE_MODE_PULSE = %100,  // two outputs, single pulse at paramx
    PWMxCFG_COMPARE_MODE_TOGGLE = %101  // two outputs, active state alternates at paramx

// default pwm config (PWMxS1CFG)
public const
    PWMxCFG_DEFAULT_CONFIG = (0<<POL2) + (0<<POL1) + PWMxCFG_LEFT_ALIGN   // out2 pol=0, out1 pol=0, left-align 

// PWM chan mask (for PWMLOAD, PWMEN)
#if (HPWM16_MAX_PWM = 1)
  const PWM_CHAN_MASK = $01
#elseif (HPWM16_MAX_PWM = 2)
  const PWM_CHAN_MASK = $03
#elseif (HPWM16_MAX_PWM = 3)
  const PWM_CHAN_MASK = $07
#elseif (HPWM16_MAX_PWM = 4)
  const PWM_CHAN_MASK = $0F
#endif
  
//------------------------------------------------------------------------------
// variables
//------------------------------------------------------------------------------
// pwm chan clock source freq, in hz
dim pwm_clock_freq(MAX_PWM) as longword

// autoload controls the LD bit function when using SetPWMx duty cycle
//  true = SetPWMx produces a LD load pulse
//  false = SetPWMx does not use LD... user must call Load()
const DEFAULT_AUTOLOAD as boolean = HPWM16_DEFAULT_AUTOLOAD 
dim autoload as boolean = DEFAULT_AUTOLOAD

//------------------------------------------------------------------------------
// functions
//------------------------------------------------------------------------------
// set autoload mode for SetPWMx()
public sub SetAutoload(load as boolean)
    autoload = load
end sub

// set PWMxCON PWM Control Register
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  bval = PWMxCON Control Register setting
public sub SetControlReg(chan as byte, bval as byte)
    select (chan)
        case 1: 
            PWM1CON = bval
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON = bval
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON = bval
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON = bval
      #endif
    end select
end sub

// enable PWM module (PWMxCON)
//  sets individual chan PWMxCON.EN bit
// parameters:
//  chan = pwm module, 1-MAX_PWM
public sub PWMEnable(chan as byte)
    select (chan)
        case 1: 
            PWM1CON.bits(EN) = 1
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON.bits(EN) = 1
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON.bits(EN) = 1
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON.bits(EN) = 1
      #endif
    end select
end sub

// disable PWM module (PWMxCON)
//  the prescaler, postscaler, and all internal logic is reset
//  outputs go to their default states, register values remain unchanged
// parameters:
//  chan = pwm module, 1-MAX_PWM
public sub PWMDisable(chan as byte)
    select (chan)
        case 1: 
            PWM1CON.bits(EN) = 0
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON.bits(EN) = 0
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON.bits(EN) = 0
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON.bits(EN) = 0
      #endif
    end select
end sub

// set multiple chan enable bits using PWMEN register
// mirror copies of all PWMxEN bits can be set simultaneously 
// to synchronize the enable event across all PWMs
// parameters:
//  chan_bits = channel bit mask  (%0000-%1111)
public inline sub PWMENMask(chan_bits as byte)
    PWMEN = chan_bits and PWM_CHAN_MASK
end sub

// load PWM duty cycle parameters (PWMxCON)
//  sets individual chan PWMxCON.LD bit
//  reload the period and duty cycle registers on the next period event
// parameters:
//  chan = pwm module, 1-MAX_PWM
public sub LoadChan(chan as byte)
    select (chan)
        case 1: 
            PWM1CON.bits(LD) = 1
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON.bits(LD) = 1
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON.bits(LD) = 1
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON.bits(LD) = 1
      #endif
    end select
end sub

// load multiple channel duty cycle parameters using PWMLOAD register
// mirror copies of all PWMxLD bits can be set simultaneously to synchronize 
// the load event across all PWMs
// parameters:
//  chan_bits = channel bit mask (%0000-%1111)
public inline sub Load(chan_bits as PWMLOAD)
end sub

// set PWMxS1CFG configuration register
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  cfg = PWMxS1CFG register setting
// note: changes to this register must be done only when the PWMxCON.EN bit is cleared
// leaves PWMxCON.EN = 0
public sub SetConfig(chan as byte, cfg as byte=PWMxCFG_DEFAULT_CONFIG)
    select (chan)
        case 1: 
            PWM1CON.bits(EN) = 0
            PWM1S1CFG = cfg
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON.bits(EN) = 0
            PWM2S1CFG = cfg
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON.bits(EN) = 0
            PWM2S1CFG = cfg
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON.bits(EN) = 0
            PWM2S1CFG = cfg
      #endif
    end select
end sub

public function GetConfig(chan as byte) as byte
    result = 0
    select (chan)
        case 1: 
            result = PWM1S1CFG
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            result = PWM2S1CFG
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            result = PWM3S1CFG
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            result = PWM4S1CFG
      #endif
    end select
end function

// set PWMxCLK clock source register (PWM timebase)
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  clk_src = clock source, PWMxCLK (defaults to FOSC)
//  clk_freq = clock source freq, in hz (if not using FOSC)
// sets PWMxCLK clock source register, leaves pwm disabled (PWMxCON.EN = 0)
// note: the hdw default setting for the PWMxCLK register is 00000, which selects the PWMIN0PPS input
public sub SetClockSrc(chan as byte, clk_src as byte=PWMxCLK_FOSC, clk_freq as longword=_clock*1000000)
    if (chan = 0) or (chan > MAX_PWM) then
        exit                // error... invalid PWM chan/module
    endif

    select (chan)
        case 1: 
            PWM1CON.bits(EN) = 0
            PWM1CLK = clk_src
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON.bits(EN) = 0
            PWM2CLK = clk_src
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON.bits(EN) = 0
            PWM3CLK = clk_src
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON.bits(EN) = 0
            PWM4CLK = clk_src
      #endif
    end select

    pwm_clock_freq(chan-1) = clk_freq
end sub

// set PWMxCPRE clock prescaler
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  prescaler = 1 (1:1) to 256 (1:256)
// the default clock prescaler is 1:1
// leaves pwm disabled (PWMxCON.EN = 0)
public sub SetPrescaler(chan as byte, prescaler as word)
    if (prescaler = 0) or (prescaler > 256) then
        exit                        // invalid prescaler value
    endif        

    // PWMxCPRE prescaler register is n=0-255 (actual prescaler is n+1)
    prescaler = prescaler - 1
    
    select (chan)
        case 1: 
            PWM1CON.bits(EN) = 0
            PWM1CPRE = prescaler.bytes(0)
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2CON.bits(EN) = 0
            PWM2CPRE = prescaler.bytes(0)
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3CON.bits(EN) = 0
            PWM3CPRE = prescaler.bytes(0)
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4CON.bits(EN) = 0
            PWM4CPRE = prescaler.bytes(0)
      #endif
    end select
end sub

// get the clock prescaler setting
// parameters:
//  chan = pwm module, 1-MAX_PWM
// returns:
//  prescaler setting, 1 (1:1) to 256 (1:256)
public function GetPrescaler(chan as byte) as word
    dim prescaler as byte = 0   // default return value

    select (chan)
        case 1: 
            prescaler = PWM1CPRE
      #if (HPWM16_MAX_PWM >= 2)
        case 2:
            prescaler = PWM2CPRE
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3:
            prescaler = PWM3CPRE
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4:
            prescaler = PWM4CPRE
      #endif
    end select

    result = prescaler + 1     // 0-255 --> 1-256
end function

// set PWMxPR period register
//  sets the number of PWM clock periods (PWMxPR+1)
//  this along with the PWM clock source and prescaler determines the PWM freq
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  period = PWMxPR period register, 0-65535
public sub SetPeriod(chan as byte, period as word)
    select (chan)
        case 1: 
            PWM1PRH = period.bytes(1)
            PWM1PRL = period.bytes(0)
      #if (HPWM16_MAX_PWM >= 2)
        case 2:
            PWM2PRH = period.bytes(1)
            PWM2PRL = period.bytes(0)
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3:
            PWM3PRH = period.bytes(1)
            PWM3PRL = period.bytes(0)
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4:
            PWM4PRH = period.bytes(1)
            PWM4PRL = period.bytes(0)
      #endif
    end select
end sub

// get the PWMxPR period register
// parameters:
//  chan = pwm module, 1-MAX_PWM
// returns:
//  PWMxPR period setting, 0-65535
public function GetPeriod(chan as byte) as word
    select (chan)
        case 1: 
            result.bytes(1) = PWM1PRH
            result.bytes(0) = PWM1PRL
      #if (HPWM16_MAX_PWM >= 2)
        case 2:
            result.bytes(1) = PWM2PRH
            result.bytes(0) = PWM2PRL
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3:
            result.bytes(1) = PWM3PRH
            result.bytes(0) = PWM3PRL
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4:
            result.bytes(1) = PWM4PRH
            result.bytes(0) = PWM4PRL
      #endif
        else
            result = 0
    end select
end function

// duty cycle settings (LEFT and RIGHT aligned modes)
//  the PWMxPR period reg determines the max values for the PWMxS1P1 and PWMxS1P2 registers
//  PWMxS1Pn = 0        output low (0%)
//  PWMxS1Pn <= PWMxPR  output duty cycle
//  PWMxS1Pn > PWMxPR   output high (100%)

// set PWMxS1P1 output parameter register
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  param1 = PWMxS1P1 parameter register setting
// note: depending on the PWMxS1CFG MODE, param1 is typically the output1 dutycycle
public sub SetPWM1(chan as byte, param1 as word)
    select (chan)
        case 1: 
            PWM1S1P1H = param1.bytes(1)
            PWM1S1P1L = param1.bytes(0)
            if (autoload) then
                PWM1CON.bits(LD) = 1
            endif
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2S1P1H = param1.bytes(1)
            PWM2S1P1L = param1.bytes(0)
            if (autoload) then
                PWM2CON.bits(LD) = 1
            endif
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3S1P1H = param1.bytes(1)
            PWM3S1P1L = param1.bytes(0)
            if (autoload) then
                PWM3CON.bits(LD) = 1
            endif
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4S1P1H = param1.bytes(1)
            PWM4S1P1L = param1.bytes(0)
            if (autoload) then
                PWM4CON.bits(LD) = 1
            endif
      #endif
    end select
end sub

// get PWMxS1P1 output parameter register
// parameters:
//  chan = pwm module, 1-MAX_PWM
public function GetPWM1(chan as byte) as word
    select (chan)
        case 1: 
            result.bytes(1) = PWM1S1P1H
            result.bytes(0) = PWM1S1P1L
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            result.bytes(1) = PWM2S1P1H
            result.bytes(0) = PWM2S1P1L
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            result.bytes(1) = PWM3S1P1H
            result.bytes(0) = PWM3S1P1L
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            result.bytes(1) = PWM4S1P1H
            result.bytes(0) = PWM4S1P1L
      #endif
        else
            result = 0
    end select
end function

// set PWMxS1P2 output parameter register
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  param2 = PWMxS1P2 parameter register setting
// note: depending on the PWMxS1CFG MODE, param2 is typically the output2 dutycycle
public sub SetPWM2(chan as byte, param2 as word, load as byte=1)
    select (chan)
        case 1: 
            PWM1S1P2H = param2.bytes(1)
            PWM1S1P2L = param2.bytes(0)
            if (autoload) then
                PWM1CON.bits(LD) = 1
            endif
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            PWM2S1P2H = param2.bytes(1)
            PWM2S1P2L = param2.bytes(0)
            if (autoload) then
                PWM2CON.bits(LD) = 1
            endif
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            PWM3S1P2H = param2.bytes(1)
            PWM3S1P2L = param2.bytes(0)
            if (autoload) then
                PWM3CON.bits(LD) = 1
            endif
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            PWM4S1P2H = param2.bytes(1)
            PWM4S1P2L = param2.bytes(0)
            if (autoload) then
                PWM4CON.bits(LD) = 1
            endif
      #endif
    end select
end sub

// get PWMxS1P2 output parameter register
// parameters:
//  chan = pwm module, 1-MAX_PWM
public function GetPWM2(chan as byte) as word
    select (chan)
        case 1: 
            result.bytes(1) = PWM1S1P2H
            result.bytes(0) = PWM1S1P2L
      #if (HPWM16_MAX_PWM >= 2)
        case 2: 
            result.bytes(1) = PWM2S1P2H
            result.bytes(0) = PWM2S1P2L
      #endif
      #if (HPWM16_MAX_PWM >= 3)
        case 3: 
            result.bytes(1) = PWM3S1P2H
            result.bytes(0) = PWM3S1P2L
      #endif
      #if (HPWM16_MAX_PWM >= 4)
        case 4: 
            result.bytes(1) = PWM4S1P2H
            result.bytes(0) = PWM4S1P2L
      #endif
        else
            result = 0
    end select
end function

// get max period value based on current setting
// parameters:
//  chan = pwm module, 1-MAX_PWM
// returns 
//  max period value, 0-65536 (PWMxPR+1)
public function MaxPeriod(chan as byte) as longword
    result = GetPeriod(chan)
    result = result + 1
end function

// given a duty cycle percentage 0-100, compute the corresponding period setting
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  pPercent = duty cycle, 0-100
// returns:
//  period register setting
public function DutyCycle(chan as byte, pPercent as byte) as word
    dim dw as longword

    if (pPercent > 100) then
        pPercent = 100
    endif
    // result = ((MaxPeriod() * pPercent) + 50) / 100
    dw = MaxPeriod(chan) * pPercent
    dw = dw + 50                   // round up
    result = dw/100
end function

// set the duty cycle for output1 as a percentage
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  pPercent = duty cycle, 0-100 percent
public sub SetDutyCycle1(chan as byte, pPercent as byte)
    dim dc as word
    
    dc = DutyCycle(chan, pPercent)
    SetPWM1(chan, dc)
end sub

// set the duty cycle for output2 as a percentage
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  pPercent = duty cycle, 0-100 percent
public sub SetDutyCycle2(chan as byte, pPercent as byte)
    dim dc as word
    
    dc = DutyCycle(chan, pPercent)
    SetPWM2(chan, dc)
end sub

//
//------------------------------------------------------------------------------
// SetFreq and related functions MaxFreq(), MinFreq(), and Freq()
//------------------------------------------------------------------------------
//
// compute PWM freq given a prescaler and period
// parameters:
//  chan = pwm module, 1-MAX_PWM
//  prescaler = prescaler setting (PWMxCPRE), 1-256
//  period = period setting (PWMxPR), 0-65535
// returns:
//  pwm freq, in hz
public function Freq(chan as byte, prescaler as word, period as word) as longword
    dim f as longword
    
    result = 0              // default return value
    if (chan = 0) or (chan > MAX_PWM) then
        exit                // error... invalid PWM chan/module
    endif

    if (prescaler = 0) then     // don't allow divide by zero
        prescaler = 1
    endif
    
    // compute PWM freq
    f = pwm_clock_freq(chan-1)  // get clock source freq
    if (f = 0) then             // pwm clock source not set...
        exit
    endif
    
    f = f / ((period + 1) * prescaler)
    if (f = 0) then             // for anything <1Hz return 1Hz
        f = 1
    endif
    
    result = f
end function

// get max pwm freq
// parameters:
//  chan = pwm module, 1-MAX_PWM
// returns:
//  max pwm freq, in hz
public function MaxFreq(chan as byte) as longword
    result = Freq(chan, 1, 1)        // PWMxCPRE prescaler = 1, PWMxPR period = 1 (50% dutycycle)
end function

// get min pwm freq
// parameters:
//  chan = pwm module, 1-MAX_PWM
// returns:
//  min pwm freq, in hz
public function MinFreq(chan as byte) as longword
    result = Freq(chan, 256, 65535)        // PWMxCPRE prescaler = 256, PWMxPR period = 65535
end function

// get current pwm freq
// parameters:
//  chan = pwm module, 1-MAX_PWM
// returns:
//  current pwm freq, in hz
// based on pwm clock, prescaler, and period register settings
public function Freq(chan as byte) as longword
    dim period as word,
        prescaler as word
    
    prescaler = GetPrescaler(chan)
    period = GetPeriod(chan)
    result = Freq(chan, prescaler, period)
end function

// set pwm freq and out1/out2 duty cycles
//  parameters:
//    chan = pwm module, 1-MAX_PWM
//    pFreq = desired PWM freq, in Hz
//    pDuty1 = out1 duty cycle in percent, 0-100
//    pDuty2 = out2 duty cycle in percent, 0-100
// if #option HPWM16_CLOSEST_FREQ = true then finds the closest PWM freq to the requested pFreq
// if autoload flag is true then PWMEnable is called to begin PWM
// exec:
// 135us-430us @ 64MHz (if HPWM16_CLOSEST_FREQ = true add 70-80us)
public function SetFreq(chan as byte, pFreq as longword, pDuty1 as byte=DEFAULT_DUTYCYCLE, pDuty2 as byte=0) as boolean
    dim prescaler as word
    dim period as word
    dim tPR as longword
    dim f as longword

  #if (HPWM16_CLOSEST_FREQ)
    dim f1 as longword
    dim df, df1 as longint
  #endif
    result = false
    
    // range check and adjust if necessary
    if (pFreq <= MinFreq(chan)) then
        prescaler = 256
        period = 65534          // reserve 65535 for 100% DC
    elseif (pFreq >= MaxFreq(chan)) then
        prescaler = 1
        period = 1
    else
        // PWM_Freq = fPWM_clock_source/((PWMxPR + 1)*prescaler)
        // solving for PWMxPR:
        //  PWMxPR  = (fPWM_clock_source/ (PWM_Freq * prescaler)) - 1
        f = pwm_clock_freq(chan-1)/pFreq    // precompute outside loop
        // get initial prescaler estimate
        // the lowest prescaler choice will give the greatest period(duty cycle) value
        prescaler = f.words(1)  // same as f/65536
        if (prescaler = 0) then
            prescaler = 1
        endif
        while (prescaler < 256)
            tPR = (f / prescaler) - 1
            if (tPR < 65535) then
                break
            endif
            prescaler = prescaler + 1
        end while
        period = tPR
        result = true
    endif
  #if (HPWM16_CLOSEST_FREQ)
    // calc resulting freq for this prescaler and period
    f = Freq(chan, prescaler, period)
    df = f - pFreq
    if (df < 0) then    // abs
        df = -df
    endif
    if (df > 0) and (period < 63334) then
        // get freq for next higher period value
        f1 = Freq(chan, prescaler, period+1)
        // determine which is closer
        df1 = f1 - pFreq
        if (df1 < 0) then   // abs
            df1 = -df1
        endif
        if (df1 < df) then     // use period+1
            period = period + 1
        endif
    endif
  #endif    // HPWM16_CLOSEST_FREQ
  
    // set pwm params
    SetPrescaler(chan, prescaler)
    SetPeriod(chan, period)
    SetDutyCycle1(chan, pDuty1)
    SetDutyCycle2(chan, pDuty2)
    if (autoload) then
        PWMEnable(chan)     // load already done by SetPWM1... just enable the pwm
    endif
end function

//------------------------------------------------------------------------------
// module initialization
//------------------------------------------------------------------------------
macro MODULE_INIT()
    clear(pwm_clock_freq)
    PWMENMask(0)      // disable all modules
#if (HPWM16_SET_DEFAULT_CLOCKSRC)
    SetClockSrc(1)
  #if (HPWM16_MAX_PWM >= 2)
    SetClockSrc(2)
  #endif
  #if (HPWM16_MAX_PWM >= 3)
    SetClockSrc(3)
  #endif
  #if (HPWM16_MAX_PWM >= 4)
    SetClockSrc(4)
  #endif
#endif      // HPWM16_SET_DEFAULT_CLOCKSRC
end macro

MODULE_INIT()

end module
